/*
 * Copyright (C) 2013 Texas Instruments Incorporated - http://www.ti.com/
 *
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include "mpm_transport_hyplnk_interface.h"
#include <ti/csl/src/ip/serdes_sb/V0/csl_wiz8_sb_refclk156p25MHz_20bit_6p25Gbps.c>
#include <ti/csl/src/ip/serdes_sb/V0/csl_wiz8_sb_refclk156p25MHz_20bit_5Gbps_sr1.c>
#include <ti/csl/src/ip/serdes_sb/V0/csl_wiz8_sb_refclk156p25MHz_20bit_10Gbps_sr1.c>
#include <ti/csl/src/ip/serdes_sb/V0/csl_wiz8_sb_refclk156p25MHz_20bit_12p5Gbps_sr1.c>
#include <ti/csl/src/ip/serdes_sb/V0/csl_wiz8_sb_refclk312p5MHz_20bit_6p25Gbps.c>
#include <ti/csl/src/ip/serdes_sb/V0/csl_wiz8_sb_refclk312p5MHz_20bit_5Gbps_sr1.c>
#include <ti/csl/src/ip/serdes_sb/V0/csl_wiz8_sb_refclk312p5MHz_20bit_10Gbps_sr1.c>
#include <ti/csl/src/ip/serdes_sb/V0/csl_wiz8_sb_refclk312p5MHz_20bit_12p5Gbps_sr1.c>

extern hyplnk_peripheral hyplnkActiveIF[NUM_INTERFACES];

int mpm_uioutil_get_string (char *filename, char *str, int str_len)
{
  FILE *fpr = 0;
  int ret_val = -1;

  if (!filename || !str || !str_len) {
    goto close_n_exit;
  }

  fpr = fopen(filename, "r");
  if (!fpr) {
    printf("Error opening file %s (%s)",
           filename, strerror(errno));
    goto close_n_exit;
  }
  if (!fgets(str, str_len, fpr)) {
    printf("Error reading file %s (%s)",
           filename, strerror(errno));
    goto close_n_exit;      
  }

  ret_val = 0;

close_n_exit:
  if (fpr) fclose(fpr);
  return ret_val;
}
char  *mpm_remove_postfix_from_device_name(char *name, int max_length)
{ 
  int index;

  index=0;
  
  /* strip anything after . in the name */
  while(name[index] != '\0')
  {
    if(name[index] == '.')
    { 
      name[index] = '\0';
      index++;
      return(&name[0]);
    }
    if(index > max_length)
      break;
    index++;
  }
  return(name);
}

int mpm_uioutil_get_device (char *uio_name, char *class_name, int class_name_length)
{
  struct dirent *entry = 0;
  char filename[MAX_FILE_NAME_LENGTH];
  char name[MAX_NAME_LENGTH];
  int ret_val = -1;
  DIR *dir = 0;
        char *name_ptr;

  if (!uio_name || !class_name || !strlen(uio_name)) {
    printf("Error in call to function");
    goto close_n_exit;
  }

  dir = opendir("/sys/class/uio");
  if (!dir) {
    printf("readdir /sys/class/uio");
    goto close_n_exit;
  }

  while ((entry = readdir(dir)) != NULL) {
    if (strstr (entry->d_name, "uio") == NULL) {
      continue;
    }

    snprintf(filename, MAX_FILE_NAME_LENGTH, "/sys/class/uio/%s/name", entry->d_name);

    if (mpm_uioutil_get_string(filename, name, MAX_NAME_LENGTH) < 0)
      goto close_n_exit;
    name_ptr = mpm_remove_postfix_from_device_name(name, MAX_NAME_LENGTH);
    if (!strncmp (uio_name, name_ptr, strlen(uio_name))) {
      ret_val = 0;
      strncpy (class_name, entry->d_name, class_name_length);
      goto close_n_exit;
    }
  }

close_n_exit:
  if (dir) closedir(dir);

  return ret_val;
}

void * internal_hyplnk_mmap(uint32_t addr, uint32_t size, int device_fd)
{
    void * virt_addr;
    uint32_t page_size, mapaddr, offset;
    page_size = getpagesize();
    mapaddr = addr & (~(page_size - 1));
    offset = addr - mapaddr;
    virt_addr = offset + mmap(0, size + offset, (PROT_READ|PROT_WRITE), MAP_SHARED, device_fd, (off_t)mapaddr);
    if ((uint32_t)virt_addr == -1)
    {
            printf("mmap failed!\n");
            return 0;
    }
    return virt_addr;	
}


void hyplnkPSCSetup(int port)
{
    int ret;
    char class_name[MAX_FILE_NAME_LENGTH];
    char dev_name[MAX_NAME_LENGTH];
    if (port == 0)
      ret = mpm_uioutil_get_device("hyperlink0", class_name, 32);
    else
      ret = mpm_uioutil_get_device("hyperlink1", class_name, 32);
    if ( ret < 0 ) {
      printf("\n Error getting device ");
    }
    snprintf(dev_name, MAX_NAME_LENGTH, "/dev/%s", class_name);

    hyplnkActiveIF[port].uio_fd = open (dev_name, (O_RDWR | O_SYNC));
    if (hyplnkActiveIF[port].uio_fd < 0)
      printf("\n Error opening device ");
    
} /* hyplnkPSCSetup */

void hyplnkPSCDisable(int port)
{
    close(hyplnkActiveIF[port].uio_fd);
} /* hyplnkPSCDisable */

int hyplnkSerdesSetup(mpm_transport_hyplnk_t *hypCfg, int index)
{
    CSL_SERDES_REF_CLOCK          refClock = CSL_SERDES_REF_CLOCK_312p5M;
    CSL_SERDES_LINK_RATE          linkRate = CSL_SERDES_LINK_RATE_6p25G;
    uint32_t                      baseAddr;
    CSL_SERDES_RESULT             csl_retval;
    uint32_t                      i;
    CSL_SERDES_LANE_CTRL_RATE     lane_rate = CSL_SERDES_LANE_FULL_RATE;
    CSL_SERDES_STATUS             pllstat;
    uint32_t serdes_regs;
    CSL_SERDES_LOOPBACK           loopback = CSL_SERDES_LOOPBACK_DISABLED;

    if (hypCfg->hyplnkRefClk == hyplnk_REFCLK_312p5)
        refClock = CSL_SERDES_REF_CLOCK_312p5M;
    else if (hypCfg->hyplnkRefClk == hyplnk_REFCLK_156p25)
        refClock = CSL_SERDES_REF_CLOCK_156p25M;
    else 
      return -1;

    if (hypCfg->hyplnkSerRate == hyplnk_SERRATE_6p25)
        linkRate = CSL_SERDES_LINK_RATE_6p25G;
    else if (hypCfg->hyplnkSerRate == hyplnk_SERRATE_12p5)
        linkRate = CSL_SERDES_LINK_RATE_12p5G;
    else 
      return -1;

    if (hypCfg->hyplnkLaneRate == hyplnk_LANE_FULL)
        lane_rate = CSL_SERDES_LANE_FULL_RATE;
    else if (hypCfg->hyplnkLaneRate == hyplnk_LANE_HALF)
        lane_rate = CSL_SERDES_LANE_HALF_RATE;
    else if (hypCfg->hyplnkLaneRate == hyplnk_LANE_QUARTER)
        lane_rate = CSL_SERDES_LANE_QUARTER_RATE;
    else 
      return -1;

    if (index == 0)
        serdes_regs = CSL_HYPERLINK_0_SERDES_CFG_REGS;
    else if (index == 1)
        serdes_regs = CSL_HYPERLINK_1_SERDES_CFG_REGS;
    else 
      return -1;

    if (hypCfg->hyplnkDir == hyplnk_LOOPBACK)
        loopback = CSL_SERDES_LOOPBACK_ENABLED;
    else if (hypCfg->hyplnkDir == hyplnk_REMOTE)
        loopback = CSL_SERDES_LOOPBACK_DISABLED;
    else 
      return -1;
    
    baseAddr = (uint32_t) internal_hyplnk_mmap(serdes_regs,0x2000, hypCfg->fd[index]);

    csl_retval = CSL_HyperlinkSerdesInit(baseAddr, refClock, linkRate);
    if (csl_retval != 0)
    {
        printf ("Invalid Serdes Init Params: %d\n", csl_retval);
    }
   
    for(i=0; i < hypCfg->hyplnkNumLanes; i++)
    {
        CSL_HyperlinkSerdesLaneEnable(baseAddr, i, loopback, lane_rate);
    }

    /* SB PLL Enable */
    CSL_HyperlinkSerdesPllEnable(baseAddr);

    /* SB PLL Status Poll */
    do
    {
        pllstat = CSL_HyperlinkSerdesGetStatus(baseAddr, hypCfg->hyplnkNumLanes);
    }while(pllstat == CSL_SERDES_STATUS_PLL_NOT_LOCKED);
   munmap((void *)baseAddr, 0x2000); 
   return 0;
}

void hyplnkSerdesDisable(mpm_transport_hyplnk_t *hypCfg, int index)
{
    uint32_t baseAddr, i, serdes_regs = 0;
    if (index == 0)
        serdes_regs = CSL_HYPERLINK_0_SERDES_CFG_REGS;
    else if (index == 1)
        serdes_regs = CSL_HYPERLINK_1_SERDES_CFG_REGS;
    baseAddr = (uint32_t) internal_hyplnk_mmap(serdes_regs,0x2000, hypCfg->fd[index]);
    CSL_HyperlinkSerdesShutdown(baseAddr);
    CSL_SerDes_CMU_Restore_Default(baseAddr);
    CSL_SerDes_COMLANE_Restore_Default(baseAddr);
    for(i=0; i < hypCfg->hyplnkNumLanes; i++)
    {
        CSL_SerDes_Lane_Restore_Default(baseAddr, i);
    }
}

hyplnkRet_e hyplnkPeripheralSetup (mpm_transport_hyplnk_t *hypCfg, int index, uint32_t timeout, int reset)
{
  hyplnkRevReg_t            rev;
  hyplnkControlReg_t        control;
  hyplnkStatusReg_t         status;
  hyplnkLinkStatusReg_t     linkStatus;
  hyplnkECCErrorsReg_t      ECCErrors;
  hyplnkLanePwrMgmtReg_t    lanePwrMgmt;
  hyplnkSERDESControl1Reg_t serdesControl1;

  hyplnkRegisters_t setRegs;
  hyplnkRegisters_t getRegs;

  hyplnkRet_e retVal;
  Hyplnk_Handle handle = NULL;
  uint32_t i = 0;

  /* To prove the APIs work -- expose uninitialized variables */
  memset (&rev,            0xff, sizeof(rev));
  memset (&control,        0xff, sizeof(control));
  memset (&status,         0xff, sizeof(status));
  memset (&linkStatus,     0xff, sizeof(linkStatus));
  memset (&ECCErrors,      0xff, sizeof(ECCErrors));
  memset (&lanePwrMgmt,    0xff, sizeof(lanePwrMgmt));
  memset (&serdesControl1, 0xff, sizeof(serdesControl1));

  /* Will always write these regs when writing */
  memset (&setRegs, 0, sizeof(setRegs));
  setRegs.control        = &control;
  setRegs.status         = &status;
  setRegs.ECCErrors      = &ECCErrors;
  setRegs.lanePwrMgmt    = &lanePwrMgmt;
  setRegs.serdesControl1 = &serdesControl1;

  /* Will always read these regs when reading */
  memset (&getRegs, 0, sizeof(getRegs));
  getRegs.control        = &control;
  getRegs.rev            = &rev;
  getRegs.status         = &status;
  getRegs.linkStatus     = &linkStatus;
  getRegs.lanePwrMgmt    = &lanePwrMgmt;
  getRegs.ECCErrors      = &ECCErrors;
  getRegs.serdesControl1 = &serdesControl1;


  if ((retVal = Hyplnk_open((int)hypCfg->hyplnkIF, &handle)) != hyplnk_RET_OK) {
    printf ("Open failed\n");
    return retVal;
  }

  /* Read rev and control */
  if ((retVal = Hyplnk_readRegs (handle, hyplnk_LOCATION_LOCAL, &getRegs)) != hyplnk_RET_OK) {
    printf ("Read revision register failed!\n");
    return retVal;
  }

  if (reset)
  {
    /* Reset the peripheral */
    control.reset       = 1;
    status.lError       = 1; /* Clear any error */
    status.rError       = 1; /* Clear any error */
    ECCErrors.sglErrCor = 0;
    ECCErrors.dblErrDet = 0;
    lanePwrMgmt.singleLane = 0;
    lanePwrMgmt.zeroLane   = 0;
    lanePwrMgmt.quadLane   = 1;
    serdesControl1.sleepCnt   = 0xff;
    serdesControl1.disableCnt = 0xff;
    if ((retVal = Hyplnk_writeRegs (handle, hyplnk_LOCATION_LOCAL, &setRegs)) != hyplnk_RET_OK) {
      printf ("[1] Control register write failed!\n");
      return retVal;
    }
    if ((retVal = Hyplnk_close (&handle)) != hyplnk_RET_OK) {
      printf ("close failed!\n");
      return retVal;
    }
    return hyplnk_RET_OK;
  }

  /* Don't need to clear any more status bits */
  setRegs.status         = NULL; 
  setRegs.ECCErrors      = NULL;
  setRegs.lanePwrMgmt    = NULL;
  setRegs.serdesControl1 = NULL;

  /* Take out of reset loopback */
  control.reset           = 0;
  control.statusIntEnable = 1; /* enable interrupt on lerror or rerror */
  control.intLocal        = 1;
  control.iLoop           = (int) hypCfg->hyplnkDir;
  if ((retVal = Hyplnk_writeRegs (handle, hyplnk_LOCATION_LOCAL, &setRegs)) != hyplnk_RET_OK) {
    printf ("[2] Control register write failed!\n");
    return retVal;
  }

  do
  {
    if (timeout == 0) i=1;
    if ((retVal = Hyplnk_readRegs (handle, hyplnk_LOCATION_LOCAL, &getRegs)) != hyplnk_RET_OK) {
      printf ("Read status failed!\n");
      return retVal;
    }
    if ((status.link) && !(status.serialHalt) && !(status.pllUnlock)) break;
    if (timeout)
    {
      usleep(1000);
      i++;
    }
    if (i==timeout) break;
  } while (1); 

  if ((retVal = Hyplnk_close (&handle)) != hyplnk_RET_OK) {
    printf ("close failed!\n");
    return retVal;
  }
  if (i==timeout)
  {
    printf("Timeout exceeded for connection, exiting...\n");
    return hyplnk_RET_INV_INITCFG;
  }
  return hyplnk_RET_OK;
}

int mpm_transport_hyplnk_init_peripheral(mpm_transport_hyplnk_t *hypCfg, int index, uint32_t timeout)
{
    hyplnkRet_e retVal;
    uint32_t dataBase_size = ((hypCfg == 0)? 0x10000000 : 0x08000000);
    hyplnkActiveIF[hypCfg->hyplnkIF].windowBase = (uint32_t) hyplnkInitCfg.dev.bases[hypCfg->hyplnkIF].dataBase;
    hyplnkActiveIF[hypCfg->hyplnkIF].configBase = (uint32_t) hyplnkInitCfg.dev.bases[hypCfg->hyplnkIF].cfgBase;
    hyplnkInitCfg.dev.bases[hypCfg->hyplnkIF].cfgBase = internal_hyplnk_mmap((uint32_t)(hyplnkInitCfg.dev.bases[hypCfg->hyplnkIF].cfgBase), 0x100, hypCfg->fd[index]);
    hyplnkInitCfg.dev.bases[hypCfg->hyplnkIF].dataBase = internal_hyplnk_mmap((uint32_t)(hyplnkInitCfg.dev.bases[hypCfg->hyplnkIF].dataBase), dataBase_size, hypCfg->fd[index]);

    if ((retVal = Hyplnk_init (&hyplnkInitCfg)) != hyplnk_RET_OK)
    {
            printf ("Hyperlink LLD device configuration failed\n");
            return (-1);
    }

    hyplnkPSCSetup((int)hypCfg->hyplnkIF);
    if ((retVal = hyplnkPeripheralSetup(hypCfg, index, timeout, 1)) != hyplnk_RET_OK)
    {
        printf("Peripheral init reset failed, retVal = %d\n",retVal);
        return -1;
    }    
    if ((hyplnkActiveIF[index].slavesActive == 1) && (hyplnkActiveIF[((~index)&0x1)].slavesActive == 0))
    {
      if (hyplnkSerdesSetup(hypCfg, 0) == -1) {
        printf("Failed to setup serdes for hyperlink 0\n");
        return -1;
      }
      if (hyplnkSerdesSetup(hypCfg, 1) == -1) {
        printf("Failed to setup serdes for hyperlink 1\n");
        return -1;
      }
    }
    if ((retVal = hyplnkPeripheralSetup(hypCfg, index, timeout, 0)) != hyplnk_RET_OK)
    {
        printf("Peripheral setup failed, retVal = %d\n",retVal);
        return -1;
    }
    return 0;
}

int mpm_transport_hyplnk_reset(mpm_transport_hyplnk_t *hypCfg, int index)
{
    hyplnkControlReg_t        control;
    hyplnkStatusReg_t         status;
    hyplnkRegisters_t setRegs;
    memset (&control, 0, sizeof(control));
    memset (&status, 0, sizeof(status));
    memset (&setRegs, 0, sizeof(setRegs));
    setRegs.control        = &control;
    setRegs.status         = &status;
    hyplnkRet_e retVal;
    Hyplnk_Handle handle = NULL;
    if (hypCfg->hyplnkIF == hyplnk_PORT_0)
    {
        hyplnkInitCfg.dev.bases[hypCfg->hyplnkIF].cfgBase = (void *)CSL_HYPERLINK_0_SLV_CFG_REGS;
        hyplnkInitCfg.dev.bases[hypCfg->hyplnkIF].dataBase = (void *)CSL_HYPERLINK_0_SLV_DATA;
    }
    else if (hypCfg->hyplnkIF == hyplnk_PORT_1)
    {
        hyplnkInitCfg.dev.bases[hypCfg->hyplnkIF].cfgBase = (void *)CSL_HYPERLINK_1_SLV_CFG_REGS;
        hyplnkInitCfg.dev.bases[hypCfg->hyplnkIF].dataBase = (void *)CSL_HYPERLINK_1_SLV_DATA;
    }

    if ((retVal = Hyplnk_open((int)hypCfg->hyplnkIF, &handle)) != hyplnk_RET_OK) {
        printf ("Open failed\n");
        return retVal;
    }
    memset (&control, 0, sizeof(control));
    do{
          /* Read rev and control */
          if ((retVal = Hyplnk_readRegs (handle, hyplnk_LOCATION_LOCAL, &setRegs)) != hyplnk_RET_OK) {
            printf ("Read revision register failed!\n");
          }
    } while(setRegs.status->rPend);

    control.serialStop = 1;
    if ((retVal = Hyplnk_writeRegs (handle, hyplnk_LOCATION_LOCAL, &setRegs)) != hyplnk_RET_OK) {
        printf ("Control register write failed!\n");
        return retVal;
    }

    memset (&control, 0, sizeof(control));
    do{
          /* Read rev and control */
          if ((retVal = Hyplnk_readRegs (handle, hyplnk_LOCATION_LOCAL, &setRegs)) != hyplnk_RET_OK) {
            printf ("Read revision register failed!\n");
          }
    } while(setRegs.status->rPend);

    control.serialStop = 0;
    control.reset = 1;
    if ((retVal = Hyplnk_writeRegs (handle, hyplnk_LOCATION_LOCAL, &setRegs)) != hyplnk_RET_OK) {
        printf ("Control register write failed!\n");
        return retVal;
    }
    hyplnkPSCDisable(index);
    Hyplnk_close(&handle);
    if ((hyplnkActiveIF[0].slavesActive == 0) && (hyplnkActiveIF[1].slavesActive == 0))
    {
      hyplnkSerdesDisable(hypCfg, 0);
      hyplnkSerdesDisable(hypCfg, 1);
    }
    return hyplnk_RET_OK;
}

void * mpm_transport_hyplnk_get_window(mpm_transport_hyplnk_t *hypCfg, uint32_t remote_addr, int index, uint32_t *offset)
{
#if TIME_PROFILE_HYPLNK_GET_WINDOW
  struct timespec tp_start, tp_end;
#endif
    hyplnkRet_e retVal;
    Hyplnk_Handle handle = NULL;
    hyplnkTXAddrOvlyReg_t TXAddrOvly;
    hyplnkRXAddrSelReg_t  RXAddrSel;
    hyplnkRXPrivIDIdxReg_t RXPrivIDIdx;
    hyplnkRXPrivIDValReg_t RXPrivIDVal;
    hyplnkRXSegIdxReg_t RXSegIdx;
    hyplnkRXSegValReg_t RXSegVal;
    hyplnkRegisters_t     localRegs;
    hyplnkRegisters_t     remoteRegs;
    uint32_t globalAddr = (uint32_t)remote_addr;
    uint32_t globalAddrBase, globalAddrOffset;
    uint32_t offsetMask = 0;
    uint32_t reg_sel_shift = 0;
    void *hlinkBase;

    memset (&localRegs,  0, sizeof(localRegs));
    memset (&remoteRegs, 0, sizeof(remoteRegs));
    memset (&TXAddrOvly, 0, sizeof(TXAddrOvly));
    memset (&RXAddrSel,  0, sizeof(RXAddrSel));
    memset (&RXPrivIDIdx,  0, sizeof(hyplnkRXPrivIDIdxReg_t));
    memset (&RXPrivIDVal,  0, sizeof(hyplnkRXPrivIDValReg_t));
    memset (&RXSegIdx,  0, sizeof(hyplnkRXSegIdxReg_t));
    memset (&RXSegVal,  0, sizeof(hyplnkRXSegValReg_t));

    localRegs.TXAddrOvly   = &TXAddrOvly;
    remoteRegs.RXAddrSel   = &RXAddrSel;
    remoteRegs.RXPrivIDIdx = &RXPrivIDIdx;
    remoteRegs.RXPrivIDVal = &RXPrivIDVal;
    remoteRegs.RXSegIdx    = &RXSegIdx;
    remoteRegs.RXSegVal    = &RXSegVal;

#if TIME_PROFILE_HYPLNK_GET_WINDOW
  clock_gettime(CLOCK_MONOTONIC, &tp_start);
#endif
    if ((retVal = Hyplnk_open((int)hypCfg->hyplnkIF, &handle)) != hyplnk_RET_OK) {
        printf ("mpm_transport_hyplnk_get_window: hyplnk open failed\n");
        return NULL;
    }
#if TIME_PROFILE_HYPLNK_GET_WINDOW
  clock_gettime(CLOCK_MONOTONIC, &tp_end);
  printf("[TIME PROFILE: mpm_transport_hyplnk_get_window] Time taken for Hyplnk_open(): %lld\n",clock_diff (&tp_start, &tp_end));
#endif

    /* This is "my side" */
    TXAddrOvly.txSecOvl    = 0;  
    TXAddrOvly.txPrivIDOvl = hypCfg->hyplnkTxPrivID; 
    TXAddrOvly.txIgnMask   = 10; /* Open 128MB window (which is common on all hyperlinks on all devices) */
  
    RXAddrSel.rxSecHi     = 0;
    RXAddrSel.rxSecLo     = 0;
    RXAddrSel.rxSecSel    = 0;
    RXAddrSel.rxPrivIDSel = hypCfg->hyplnkRxPrivID; /* Symmetric with TXAddrOvly.txPrivIDOvl */
    RXAddrSel.rxSegSel    = hypCfg->hyplnkSegSel;
  
    if (hypCfg->hyplnkSegSel > 0)
      offsetMask = (1<< (hypCfg->hyplnkSegSel+16)) - 1;
    else
      offsetMask = 0xFFFFFFFF;
    
    globalAddrBase   = globalAddr & ~offsetMask;
    if (globalAddrBase == 0)
    {
      globalAddrBase = globalAddr;
      globalAddrOffset = globalAddr & 0xFFFF;
    }
    else
      globalAddrOffset = globalAddr &  offsetMask;

#if TIME_PROFILE_HYPLNK_GET_WINDOW
  clock_gettime(CLOCK_MONOTONIC, &tp_start);
#endif
    if ((retVal = Hyplnk_getWindow (handle, &hlinkBase, NULL)) != hyplnk_RET_OK) {
        printf("hyplnk_getWindow failed\n");
        return NULL;
    }
#if TIME_PROFILE_HYPLNK_GET_WINDOW
  clock_gettime(CLOCK_MONOTONIC, &tp_end);
  printf("[TIME PROFILE: mpm_transport_hyplnk_get_window] Time taken for Hyplnk_getWindow(): %lld\n",clock_diff (&tp_start, &tp_end));
#endif

#if TIME_PROFILE_HYPLNK_GET_WINDOW
  clock_gettime(CLOCK_MONOTONIC, &tp_start);
#endif
    if ((retVal = Hyplnk_writeRegs (handle, hyplnk_LOCATION_LOCAL, &localRegs)) != hyplnk_RET_OK) {
        printf ("mpm_transport_hyplnk_get_window: Local register write failed!\n");
        return NULL;
    }
#if TIME_PROFILE_HYPLNK_GET_WINDOW
  clock_gettime(CLOCK_MONOTONIC, &tp_end);
  printf("[TIME PROFILE: mpm_transport_hyplnk_get_window] Time taken for Hyplnk_writeRegs() to Local: %lld\n",clock_diff (&tp_start, &tp_end));
#endif

  RXPrivIDIdx.rxPrivIDIdx = index;
  RXSegIdx.rxSegIdx = index;
#if TIME_PROFILE_HYPLNK_GET_WINDOW
  clock_gettime(CLOCK_MONOTONIC, &tp_start);
#endif
    if ((retVal = Hyplnk_writeRegs (handle, hyplnk_LOCATION_REMOTE, &remoteRegs)) != hyplnk_RET_OK) {
        printf ("mpm_transport_hyplnk_get_window: Remote register write failed!\n");
        return NULL;
    }
#if TIME_PROFILE_HYPLNK_GET_WINDOW
  clock_gettime(CLOCK_MONOTONIC, &tp_end);
  printf("[TIME PROFILE: mpm_transport_hyplnk_get_window] Time taken for Hyplnk_writeRegs() to Remote: %lld\n",clock_diff (&tp_start, &tp_end));
#endif

  RXPrivIDVal.rxPrivIDVal = index;
  RXSegVal.rxLenVal = hypCfg->hyplnkRxLen;
  RXSegVal.rxSegVal = globalAddrBase >> 16;
#if TIME_PROFILE_HYPLNK_GET_WINDOW
  clock_gettime(CLOCK_MONOTONIC, &tp_start);
#endif
    if ((retVal = Hyplnk_writeRegs (handle, hyplnk_LOCATION_REMOTE, &remoteRegs)) != hyplnk_RET_OK) {
        printf ("mpm_transport_hyplnk_get_window: Remote register write failed!\n");
        return NULL;
    }
#if TIME_PROFILE_HYPLNK_GET_WINDOW
  clock_gettime(CLOCK_MONOTONIC, &tp_end);
  printf("[TIME PROFILE: mpm_transport_hyplnk_get_window] Time taken for Hyplnk_writeRegs() to Remote: %lld\n",clock_diff (&tp_start, &tp_end));
#endif

#if TIME_PROFILE_HYPLNK_GET_WINDOW
  clock_gettime(CLOCK_MONOTONIC, &tp_start);
#endif
    if ((retVal = Hyplnk_close (&handle)) != hyplnk_RET_OK) {
        printf ("mpm_transport_hyplnk_get_window: close failed!\n");
        return NULL;
    }
#if TIME_PROFILE_HYPLNK_GET_WINDOW
  clock_gettime(CLOCK_MONOTONIC, &tp_end);
  printf("[TIME PROFILE: mpm_transport_hyplnk_get_window] Time taken for Hyplnk_close(): %lld\n",clock_diff (&tp_start, &tp_end));
#endif

    if (hypCfg->hyplnkSegSel)
    {
        reg_sel_shift = index << (hypCfg->hyplnkSegSel + 16);
    }
    if (reg_sel_shift >= 0x10000000)
    {
      printf("Index out of bounds, please use smaller segment sizes and a lower rxseg value\n");
      return NULL;
    }
    *offset = globalAddrOffset + reg_sel_shift;

    debug_printf("index is 0x%08x\n",index);
    debug_printf("globalAddr is 0x%08x\n", globalAddr);
    debug_printf("globalAddrBase is 0x%08x\n", globalAddrBase);
    debug_printf("globalAddrOffset is 0x%08x\n", globalAddrOffset);
    debug_printf("reg_sel_shift is 0x%08x\n", reg_sel_shift);
    return (void *)( (char *)hlinkBase);
}
